@@ -1,6 +1,10 @@
 from enum import Enum
 
 
+class ReversedEnum(Enum):
+    ...
+
+
 class MessageType(Enum):
     operation = 'operation'
     big_map = 'big_map'
@@ -10,6 +14,7 @@ class MessageType(Enum):
 class IndexType(Enum):
     operation = 'operation'
     big_map = 'big_map'
+    head = 'head'
 
 
 class IndexStatus(Enum):
@@ -20,10 +25,9 @@ class IndexStatus(Enum):
     ONESHOT = 'ONESHOT'
 
 
-class ReindexingReason(Enum):
+class ReindexingReason(ReversedEnum):
     MANUAL = 'triggered manually from callback'
     MIGRATION = 'applied migration requires reindexing'
-    CLI_OPTION = 'run with `--reindex` option'
     ROLLBACK = 'reorg message received and can\'t be processed'
     CONFIG_HASH_MISMATCH = 'index config has been modified'
     SCHEMA_HASH_MISMATCH = 'database schema has been modified'
