@@ -1,6 +1,6 @@
 from abc import abstractmethod
 from collections import defaultdict, deque, namedtuple
-from typing import Deque, Dict, List, Optional, Set, Tuple, Union, cast
+from typing import Deque, Dict, Iterable, Optional, Sequence, Set, Tuple, Union, cast
 
 from pydantic.error_wrappers import ValidationError
 
@@ -9,6 +9,8 @@ from dipdup.config import (
     BigMapHandlerConfig,
     BigMapIndexConfig,
     ContractConfig,
+    HeadHandlerConfig,
+    HeadIndexConfig,
     OperationHandlerConfig,
     OperationHandlerOriginationPatternConfig,
     OperationHandlerPatternConfigT,
@@ -20,17 +22,31 @@ from dipdup.config import (
 from dipdup.context import DipDupContext
 from dipdup.datasources.tzkt.datasource import BigMapFetcher, OperationFetcher, TzktDatasource
 from dipdup.exceptions import ConfigInitializationException, InvalidDataError, ReindexingReason
-from dipdup.models import BigMapData, BigMapDiff, BlockData, IndexStatus, OperationData, Origination, Transaction
+from dipdup.models import BigMapData, BigMapDiff, BlockData, HeadBlockData, IndexStatus, OperationData, Origination, Transaction
 from dipdup.utils import FormattedLogger
 from dipdup.utils.database import in_global_transaction
 
 # NOTE: Operations of a single contract call
 OperationSubgroup = namedtuple('OperationSubgroup', ('hash', 'counter'))
 
-_cached_blocks: Dict[int, BlockData] = {}
+# NOTE: Message queue of OperationIndex
+SingleLevelRollback = namedtuple('SingleLevelRollback', ('level'))
+Operations = Tuple[OperationData, ...]
+OperationQueueItemT = Union[Operations, SingleLevelRollback]
+OperationHandlerArgumentT = Optional[Union[Transaction, Origination, OperationData]]
+MatchedOperationsT = Tuple[OperationSubgroup, OperationHandlerConfig, Deque[OperationHandlerArgumentT]]
+MatchedBigMapsT = Tuple[BigMapHandlerConfig, BigMapDiff]
+
+# NOTE: For initializing the index state on startup
+block_cache: Dict[int, BlockData] = {}
 
 
 class Index:
+    """Base class for index implementations
+
+    Provides common interface for managing index state and switching between sync and realtime modes.
+    """
+
     _queue: Deque
 
     def __init__(self, ctx: DipDupContext, config: ResolvedIndexConfigT, datasource: TzktDatasource) -> None:
@@ -55,11 +71,16 @@ class Index:
         if self._state:
             raise RuntimeError('Index state is already initialized')
 
+        if isinstance(self._config, (OperationIndexConfig, BigMapIndexConfig)) and self._config.first_level:
+            level = self._config.first_level
+        else:
+            level = 0
+
         self._state, created = await models.Index.get_or_create(
             name=self._config.name,
             type=self._config.kind,
             defaults=dict(
-                level=self._config.first_level,
+                level=level,
                 config_hash=self._config.hash(),
                 template=self._config.parent.name if self._config.parent else None,
                 template_values=self._config.template_values,
@@ -73,14 +94,14 @@ class Index:
         if not head:
             return
 
-        if head.level not in _cached_blocks:
-            _cached_blocks[head.level] = await self.datasource.get_block(head.level)
-        if head.hash != _cached_blocks[head.level].hash:
+        if head.level not in block_cache:
+            block_cache[head.level] = await self.datasource.get_block(head.level)
+        if head.hash != block_cache[head.level].hash:
             await self._ctx.reindex(ReindexingReason.BLOCK_HASH_MISMATCH)
 
     async def process(self) -> None:
         # NOTE: `--oneshot` flag implied
-        if self._config.last_level:
+        if isinstance(self._config, (OperationIndexConfig, BigMapIndexConfig)) and self._config.last_level:
             last_level = self._config.last_level
             await self._synchronize(last_level, cache=True)
             await self.state.update_status(IndexStatus.ONESHOT, last_level)
@@ -89,7 +110,9 @@ class Index:
             raise RuntimeError('Call `set_sync_level` before starting IndexDispatcher')
 
         elif self.state.level < self._datasource.sync_level:
-            self._logger.info('Index is behind datasource, sync to datasource level')
+            self._logger.info(
+                'Index is behind datasource, sync to datasource level: %s -> %s', self.state.level, self._datasource.sync_level
+            )
             self._queue.clear()
             last_level = self._datasource.sync_level
             await self._synchronize(last_level)
@@ -124,44 +147,57 @@ class Index:
         self._logger.info('Index is synchronized to level %s', last_level)
         await self.state.update_status(status=IndexStatus.REALTIME, level=last_level)
 
+    def _extract_level(self, message: Union[Tuple[OperationData, ...], Tuple[BigMapData, ...]]) -> int:
+        batch_levels = tuple(set(item.level for item in message))
+        if len(batch_levels) != 1:
+            raise RuntimeError(f'Items in operation/big_map batch have different levels: {batch_levels}')
+        return tuple(batch_levels)[0]
+
 
 class OperationIndex(Index):
     _config: OperationIndexConfig
 
     def __init__(self, ctx: DipDupContext, config: OperationIndexConfig, datasource: TzktDatasource) -> None:
         super().__init__(ctx, config, datasource)
-        self._queue: Deque[Tuple[int, List[OperationData]]] = deque()
+        self._queue: Deque[OperationQueueItemT] = deque()
         self._contract_hashes: Dict[str, Tuple[int, int]] = {}
         self._rollback_level: Optional[int] = None
-        self._last_hashes: Set[str] = set()
+        self._head_hashes: Set[str] = set()
         self._migration_originations: Optional[Dict[str, OperationData]] = None
 
-    def push(self, level: int, operations: List[OperationData]) -> None:
-        self._queue.append((level, operations))
+    def push_operations(self, operations: Tuple[OperationData, ...]) -> None:
+        self._queue.append(operations)
 
-    async def single_level_rollback(self, from_level: int) -> None:
-        """Ensure next arrived block is the same as rolled back one
+    def push_rollback(self, level: int) -> None:
+        self._queue.append(SingleLevelRollback(level))
 
-        Called by IndexDispatcher in case index datasource reported a rollback.
+    async def _single_level_rollback(self, level: int) -> None:
+        """Ensure next arrived block has all operations of the previous block. But it could also contain additional operations.
+
+        Called by IndexDispatcher when index datasource receive a single level rollback.
         """
         if self._rollback_level:
-            raise RuntimeError('Already in rollback state')
+            raise RuntimeError('Index is already in rollback state')
 
-        if self.state.level < from_level:
-            self._logger.info('Index level is lower than rollback level, ignoring')
-        elif self.state.level == from_level:
-            self._logger.info('Single level rollback has been triggered')
-            self._rollback_level = from_level
+        state_level = cast(int, self.state.level)
+        if state_level < level:
+            self._logger.info('Index level is lower than rollback level, ignoring: %s < %s', state_level, level)
+        elif state_level == level:
+            self._logger.info('Single level rollback, next block will be processed partially')
+            self._rollback_level = level
         else:
-            raise RuntimeError('Index level is higher than rollback level')
+            raise RuntimeError(f'Index level is higher than rollback level: {state_level} > {level}')
 
     async def _process_queue(self) -> None:
         """Process WebSocket queue"""
-        if self._queue:
-            self._logger.info('Processing websocket queue')
         while self._queue:
-            level, operations = self._queue.popleft()
-            await self._process_level_operations(level, operations)
+            message = self._queue.popleft()
+            if isinstance(message, SingleLevelRollback):
+                self._logger.info('Processing rollback realtime message, %s left in queue', len(self._queue))
+                await self._single_level_rollback(message.level)
+            else:
+                self._logger.info('Processing operations realtime message, %s left in queue', len(self._queue))
+                await self._process_level_operations(message)
 
     async def _synchronize(self, last_level: int, cache: bool = False) -> None:
         """Fetch operations via Fetcher and pass to message callback"""
@@ -173,9 +209,9 @@ class OperationIndex(Index):
         transaction_addresses = await self._get_transaction_addresses()
         origination_addresses = await self._get_origination_addresses()
 
-        migration_originations = []
+        migration_originations: Tuple[OperationData, ...] = ()
         if self._config.types and OperationType.migration in self._config.types:
-            migration_originations = await self._datasource.get_migration_originations(first_level)
+            migration_originations = tuple(await self._datasource.get_migration_originations(first_level))
             for op in migration_originations:
                 code_hash, type_hash = await self._get_contract_hashes(cast(str, op.originated_contract_address))
                 op.originated_contract_code_hash, op.originated_contract_type_hash = code_hash, type_hash
@@ -190,14 +226,15 @@ class OperationIndex(Index):
             migration_originations=migration_originations,
         )
 
-        async for level, operations in fetcher.fetch_operations_by_level():
-            await self._process_level_operations(level, operations)
+        async for _, operations in fetcher.fetch_operations_by_level():
+            await self._process_level_operations(operations)
 
         await self._exit_sync_state(last_level)
 
-    async def _process_level_operations(self, level: int, operations: List[OperationData]) -> None:
-        if level <= self.state.level:
-            raise RuntimeError(f'Level of operation batch must be higher than index state level: {level} <= {self.state.level}')
+    async def _process_level_operations(self, operations: Tuple[OperationData, ...]) -> None:
+        if not operations:
+            return
+        level = self._extract_level(operations)
 
         if self._rollback_level:
             levels = {
@@ -210,23 +247,36 @@ class OperationIndex(Index):
                 raise RuntimeError(f'Index is in a rollback state, but received operation batch with different levels: {levels_repr}')
 
             self._logger.info('Rolling back to previous level, verifying processed operations')
-            expected_hashes = set(self._last_hashes)
-            received_hashes = set([op.hash for op in operations])
-            reused_hashes = received_hashes & expected_hashes
-            if reused_hashes != expected_hashes:
+            expected_hashes = set(self._head_hashes)
+            received_hashes = set(op.hash for op in operations)
+            new_hashes = received_hashes - expected_hashes
+            missing_hashes = expected_hashes - received_hashes
+
+            self._logger.info('Comparing hashes: %s new, %s missing', len(new_hashes), len(missing_hashes))
+            if missing_hashes:
+                self._logger.info('Some operations are backtracked: %s', ', '.join(missing_hashes))
                 await self._ctx.reindex(ReindexingReason.ROLLBACK)
 
             self._rollback_level = None
-            self._last_hashes = set()
-            new_hashes = received_hashes - expected_hashes
-            if not new_hashes:
-                return
-            operations = [op for op in operations if op.hash in new_hashes]
+            self._head_hashes = set()
+            operations = tuple(op for op in operations if op.hash in new_hashes)
+
+        # NOTE: le operator because it could be a single level rollback
+        elif level < self.state.level:
+            raise RuntimeError(f'Level of operation batch must be higher than index state level: {level} < {self.state.level}')
+
+        self._logger.info('Processing %s operations of level %s', len(operations), level)
+        matched_subgroups = await self._match_operations(operations)
+
+        # NOTE: We still need to bump index level but don't care if it will be done in existing transaction
+        if not matched_subgroups:
+            await self.state.update_status(level=level)
+            return
 
         async with in_global_transaction():
-            self._logger.info('Processing %s operations of level %s', len(operations), level)
-            await self._process_operations(operations)
-            await self.state.update_status(self.state.status, level)
+            for operation_subgroup, handler_config, args in matched_subgroups:
+                await self._call_matched_handler(handler_config, operation_subgroup, args)
+            await self.state.update_status(level=level)
 
     async def _match_operation(self, pattern_config: OperationHandlerPatternConfigT, operation: OperationData) -> bool:
         """Match single operation with pattern"""
@@ -261,14 +311,15 @@ class OperationIndex(Index):
         else:
             raise NotImplementedError
 
-    async def _process_operations(self, operations: List[OperationData]) -> None:
+    async def _match_operations(self, operations: Iterable[OperationData]) -> Deque[MatchedOperationsT]:
         """Try to match operations in cache with all patterns from indexes. Must be wrapped in transaction."""
-        self._last_hashes = set()
-        operation_subgroups: Dict[OperationSubgroup, List[OperationData]] = defaultdict(list)
+        self._head_hashes.clear()
+        matched_subgroups: Deque[MatchedOperationsT] = deque()
+        operation_subgroups: Dict[OperationSubgroup, Deque[OperationData]] = defaultdict(deque)
         for operation in operations:
             key = OperationSubgroup(operation.hash, operation.counter)
             operation_subgroups[key].append(operation)
-            self._last_hashes.add(operation.hash)
+            self._head_hashes.add(operation.hash)
 
         for operation_subgroup, operations in operation_subgroups.items():
             self._logger.debug('Matching %s', key)
@@ -276,9 +327,9 @@ class OperationIndex(Index):
             for handler_config in self._config.handlers:
                 operation_idx = 0
                 pattern_idx = 0
-                matched_operations: List[Optional[OperationData]] = []
+                matched_operations: Deque[Optional[OperationData]] = deque()
 
-                # TODO: Ensure complex cases work, for ex. required argument after optional one
+                # TODO: Ensure complex cases work, e.g. when optional argument is followed by required one
                 # TODO: Add None to matched_operations where applicable (pattern is optional and operation not found)
                 while operation_idx < len(operations):
                     operation, pattern_config = operations[operation_idx], handler_config.pattern[pattern_idx]
@@ -302,26 +353,29 @@ class OperationIndex(Index):
                         operation_idx += 1
 
                     if pattern_idx == len(handler_config.pattern):
-                        await self._on_match(operation_subgroup, handler_config, matched_operations)
+                        self._logger.info('%s: `%s` handler matched!', operation_subgroup.hash, handler_config.callback)
 
-                        matched_operations = []
+                        args = await self._prepare_handler_args(handler_config, matched_operations)
+                        matched_subgroups.append((operation_subgroup, handler_config, args))
+
+                        matched_operations.clear()
                         pattern_idx = 0
 
                 if len(matched_operations) >= sum(map(lambda x: 0 if x.optional else 1, handler_config.pattern)):
-                    await self._on_match(operation_subgroup, handler_config, matched_operations)
+                    self._logger.info('%s: `%s` handler matched!', operation_subgroup.hash, handler_config.callback)
 
-    async def _on_match(
+                    args = await self._prepare_handler_args(handler_config, matched_operations)
+                    matched_subgroups.append((operation_subgroup, handler_config, args))
+
+        return matched_subgroups
+
+    async def _prepare_handler_args(
         self,
-        operation_subgroup: OperationSubgroup,
         handler_config: OperationHandlerConfig,
-        matched_operations: List[Optional[OperationData]],
-    ):
-        """Prepare handler arguments, parse parameter and storage. Schedule callback in executor."""
-        self._logger.info('%s: `%s` handler matched!', operation_subgroup.hash, handler_config.callback)
-        if not handler_config.parent:
-            raise ConfigInitializationException
-
-        args: List[Optional[Union[Transaction, Origination, OperationData]]] = []
+        matched_operations: Deque[Optional[OperationData]],
+    ) -> Deque[OperationHandlerArgumentT]:
+        """Prepare handler arguments, parse parameter and storage."""
+        args: Deque[OperationHandlerArgumentT] = deque()
         for pattern_config, operation in zip(handler_config.pattern, matched_operations):
             if operation is None:
                 args.append(None)
@@ -360,6 +414,14 @@ class OperationIndex(Index):
             else:
                 raise NotImplementedError
 
+        return args
+
+    async def _call_matched_handler(
+        self, handler_config: OperationHandlerConfig, operation_subgroup: OperationSubgroup, args: Sequence[OperationHandlerArgumentT]
+    ) -> None:
+        if not handler_config.parent:
+            raise ConfigInitializationException
+
         await self._ctx.fire_handler(
             handler_config.callback,
             handler_config.parent.name,
@@ -405,18 +467,18 @@ class BigMapIndex(Index):
 
     def __init__(self, ctx: DipDupContext, config: BigMapIndexConfig, datasource: TzktDatasource) -> None:
         super().__init__(ctx, config, datasource)
-        self._queue: Deque[Tuple[int, List[BigMapData]]] = deque()
+        self._queue: Deque[Tuple[BigMapData, ...]] = deque()
 
-    def push(self, level: int, big_maps: List[BigMapData]):
-        self._queue.append((level, big_maps))
+    def push_big_maps(self, big_maps: Tuple[BigMapData, ...]) -> None:
+        self._queue.append(big_maps)
 
     async def _process_queue(self) -> None:
         """Process WebSocket queue"""
         if self._queue:
             self._logger.info('Processing websocket queue')
         while self._queue:
-            level, big_maps = self._queue.popleft()
-            await self._process_level_big_maps(level, big_maps)
+            big_maps = self._queue.popleft()
+            await self._process_level_big_maps(big_maps)
 
     async def _synchronize(self, last_level: int, cache: bool = False) -> None:
         """Fetch operations via Fetcher and pass to message callback"""
@@ -438,18 +500,31 @@ class BigMapIndex(Index):
             cache=cache,
         )
 
-        async for level, big_maps in fetcher.fetch_big_maps_by_level():
-            await self._process_level_big_maps(level, big_maps)
+        async for _, big_maps in fetcher.fetch_big_maps_by_level():
+            await self._process_level_big_maps(big_maps)
 
         await self._exit_sync_state(last_level)
 
-    async def _process_level_big_maps(self, level: int, big_maps: List[BigMapData]):
+    async def _process_level_big_maps(self, big_maps: Tuple[BigMapData, ...]):
+        if not big_maps:
+            return
+        level = self._extract_level(big_maps)
+
+        # NOTE: le operator because single level rollbacks are not supported
         if level <= self.state.level:
             raise RuntimeError(f'Level of big map batch must be higher than index state level: {level} <= {self.state.level}')
 
+        self._logger.info('Processing %s big map diffs of level %s', len(big_maps), level)
+        matched_big_maps = await self._match_big_maps(big_maps)
+
+        # NOTE: We still need to bump index level but don't care if it will be done in existing transaction
+        if not matched_big_maps:
+            await self.state.update_status(level=level)
+            return
+
         async with in_global_transaction():
-            self._logger.info('Processing %s big map diffs of level %s', len(big_maps), level)
-            await self._process_big_maps(big_maps)
+            for handler_config, big_map_diff in matched_big_maps:
+                await self._call_matched_handler(handler_config, big_map_diff)
             await self.state.update_status(level=level)
 
     async def _match_big_map(self, handler_config: BigMapHandlerConfig, big_map: BigMapData) -> bool:
@@ -460,11 +535,11 @@ class BigMapIndex(Index):
             return False
         return True
 
-    async def _on_match(
+    async def _prepare_handler_args(
         self,
         handler_config: BigMapHandlerConfig,
         matched_big_map: BigMapData,
-    ) -> None:
+    ) -> BigMapDiff:
         """Prepare handler arguments, parse key and value. Schedule callback in executor."""
         self._logger.info('%s: `%s` handler matched!', matched_big_map.operation_id, handler_config.callback)
         if not handler_config.parent:
@@ -488,13 +563,30 @@ class BigMapIndex(Index):
         else:
             value = None
 
-        big_map_diff = BigMapDiff(  # type: ignore
+        return BigMapDiff(
             data=matched_big_map,
             action=matched_big_map.action,
             key=key,
             value=value,
         )
 
+    async def _match_big_maps(self, big_maps: Iterable[BigMapData]) -> Deque[MatchedBigMapsT]:
+        """Try to match big map diffs in cache with all patterns from indexes."""
+        matched_big_maps: Deque[MatchedBigMapsT] = deque()
+
+        for big_map in big_maps:
+            for handler_config in self._config.handlers:
+                big_map_matched = await self._match_big_map(handler_config, big_map)
+                if big_map_matched:
+                    arg = await self._prepare_handler_args(handler_config, big_map)
+                    matched_big_maps.append((handler_config, arg))
+
+        return matched_big_maps
+
+    async def _call_matched_handler(self, handler_config: BigMapHandlerConfig, big_map_diff: BigMapDiff) -> None:
+        if not handler_config.parent:
+            raise ConfigInitializationException
+
         await self._ctx.fire_handler(
             handler_config.callback,
             handler_config.parent.name,
@@ -504,15 +596,6 @@ class BigMapIndex(Index):
             big_map_diff,
         )
 
-    async def _process_big_maps(self, big_maps: List[BigMapData]) -> None:
-        """Try to match big map diffs in cache with all patterns from indexes."""
-
-        for big_map in big_maps:
-            for handler_config in self._config.handlers:
-                big_map_matched = await self._match_big_map(handler_config, big_map)
-                if big_map_matched:
-                    await self._on_match(handler_config, big_map)
-
     async def _get_big_map_addresses(self) -> Set[str]:
         """Get addresses to fetch big map diffs from during initial synchronization"""
         addresses = set()
@@ -526,3 +609,45 @@ class BigMapIndex(Index):
         for handler_config in self._config.handlers:
             paths.add(handler_config.path)
         return paths
+
+
+class HeadIndex(Index):
+    _config: HeadIndexConfig
+
+    def __init__(self, ctx: DipDupContext, config: HeadIndexConfig, datasource: TzktDatasource) -> None:
+        super().__init__(ctx, config, datasource)
+        self._queue: Deque[HeadBlockData] = deque()
+
+    async def _synchronize(self, last_level: int, cache: bool = False) -> None:
+        self._logger.info('Setting index level to %s and moving on', last_level)
+        await self.state.update_status(status=IndexStatus.REALTIME, level=last_level)
+
+    async def _process_queue(self) -> None:
+        while self._queue:
+            head = self._queue.popleft()
+            self._logger.info('Processing head realtime message, %s left in queue', len(self._queue))
+
+            level = head.level
+            if level <= self.state.level:
+                raise RuntimeError(f'Level of head must be higher than index state level: {level} <= {self.state.level}')
+
+            async with in_global_transaction():
+                self._logger.info('Processing head info of level %s', level)
+                for handler_config in self._config.handlers:
+                    await self._call_matched_handler(handler_config, head)
+                await self.state.update_status(level=level)
+
+    async def _call_matched_handler(self, handler_config: HeadHandlerConfig, head: HeadBlockData) -> None:
+        if not handler_config.parent:
+            raise ConfigInitializationException
+
+        await self._ctx.fire_handler(
+            handler_config.callback,
+            handler_config.parent.name,
+            self.datasource,
+            head.hash,
+            (head,),
+        )
+
+    def push_head(self, head: HeadBlockData) -> None:
+        self._queue.append(head)
